/***************************************************************/
/*     Cable Exercise for CMACS workshop                       */
/*                                                             */
/*                                                             */
/*     Author: E. Bartocci (eziobart@ams.sunysb.edu)           */
/*                                                             */
/*     Description:                                            */
/*                                                             */
/*        We provide the code for the simulation of a cable    */
/*     of cells implementing the Minimal Model of Flavio Fenton*/
/*     The simulation can be performed both in a serial way and*/
/*     parallel way using CUDA (Commenting the CUDA_ENABLED).  */
/*                                                             */ 
/*     Date:  11/29/10                                         */
/*                                                             */
/*   Free distribution with authors permission                 */
/*                                                             */
/* ==   SUNY Stony Brook, Stony Brook, NY                      */
/* ==                                                          */ 
/* ============================================================*/   


/*The following are the parameters that you can find in the paper
 * A. Bueno-Orovio, M. Cherry, and F. Fenton, Minimal model for 
 * human ventricular action potentials in tissue, Journal of 
 * Theoretical Biology, no. 253, pp. 544560, 2008.              */

#define EPI_TVP        1.4506
#define ONE_O_EPI_TVP  0.68936991589687

#define EPI_TV1M     60.
#define EPI_TV2M     1150.

#define ONE_O_EPI_TV1M 0.016666666666667
#define ONE_O_EPI_TV2M 0.000869565217391

#define EPI_TWP      200.0

#define ONE_O_EPI_TWP  0.005

#define EPI_TW1M     60.0
#define EPI_TW2M     15.     //The same with Flavio's paper

#define EPI_TS1      2.7342  //The same with Flavio's paper
#define EPI_TS2      16.     //The same with Flavio's paper

#define ONE_O_EPI_TS1 0.365737692926633
#define ONE_O_EPI_TS2 0.0625

#define EPI_TFI      0.11    //The same with Flavio's paper
#define ONE_O_EPI_TFI 9.090909090909091

#define EPI_TO1      400.    //The same with Flavio's paper
#define EPI_TO2      6.      //The same with Flavio's paper

#define ONE_O_EPI_TO1 0.0025
#define ONE_O_EPI_TO2 0.166666666666667

#define ONE_O_EPI_TO1 0.0025
#define ONE_O_EPI_TO2 0.166666666666667

#define EPI_TSO1     30.0181 //The same with Flavio's paper
#define EPI_TSO2     0.9957  //The same with Flavio's paper

#define EPI_TSI      1.8875
#define ONE_O_EPI_TSI 0.529801324503311

#define EPI_TWINF    0.07
#define ONE_O_EPI_TWINF    14.285714285714286

#define EPI_THV      0.3
#define EPI_THVM     0.006
#define EPI_THVINF   0.006
#define EPI_THW      0.13
#define EPI_THWINF   0.006
#define EPI_THSO     0.13
#define EPI_THSI     0.13
#define EPI_THO      0.006
#define EPI_KWM      65.     //The same of Flavio's paper
#define EPI_KS       2.0994  //The same of Flavio's paper
#define EPI_KSO      2.0458  //The same of Flavio's paper
#define EPI_UWM      0.03    //The same of Flavio's paper
#define EPI_US       0.9087  //The same of Flavio's paper
#define EPI_UO       0.     // The same of Flavio's paper
#define EPI_UU       1.55   // The same of Flavio's paper
#define EPI_USO      0.65   // The same of Flavio's paper
#define EPI_SC       0.007

#define EPI_WINFSTAR 0.94   // The same of Flavio's paper

#define EPI_TW2M_TW1M -45.0
#define EPI_TSO2_TSO1 -29.0224

#define EPI_TW2M_TW1M_DIVBY2 -22.5
#define EPI_TSO2_TSO1_DIVBY2 -14.5112


typedef float Real; /*To use double we need a NVIDIA card
                      with CUDA compute capability 1.3*/

typedef struct __align__(4*sizeof(Real)){
    Real u, v, w, s;
} MMCell;




#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "book.h"
#include <cutil_inline.h>





__global__ void  cstep_epi_gpu_aligned ( MMCell *cable, int nxb, Real dt, Real *lap, Real stim){


	  Real u, v, w, s;
	  Real tvm, vinf, winf;
	  Real jfi, jso, jsi;
	  Real twm, tso, ts, to, ds, dw, dv;
	  
	  
	  int i = blockIdx.x * blockDim.x + threadIdx.x;
	  if (i >= 0 && i < nxb-1){
		  
		  u = cable[i].u;
		  v = cable[i].v;
		  w = cable[i].w;
		  s = cable[i].s;
		  
		  
	      tvm =  (u > EPI_THVM) ? ONE_O_EPI_TV2M : ONE_O_EPI_TV1M;
          tso =  1./(EPI_TSO1 + (EPI_TSO2_TSO1_DIVBY2)*(1.+tanh(EPI_KSO*(u-EPI_USO))));
          twm =  1./(EPI_TW1M + (EPI_TW2M_TW1M_DIVBY2)*(1.+tanh(EPI_KWM*(u-EPI_UWM))));   

          ts  = (u  > EPI_THW)  ? ONE_O_EPI_TS2  : ONE_O_EPI_TS1;
          to  = (u  > EPI_THO)  ? ONE_O_EPI_TO2  : ONE_O_EPI_TO1;

          vinf = (u  >  EPI_THVINF) ? 0.0: 1.0;
	      winf = (u  >  EPI_THWINF) ? EPI_WINFSTAR: 1.0- u * ONE_O_EPI_TWINF;

           dv = (u  > EPI_THV) ? -v*ONE_O_EPI_TVP : (vinf-v)*tvm;
           dw = (u  > EPI_THW) ? -w*ONE_O_EPI_TWP : (winf-w)*twm;
           ds = (((1.+tanh(EPI_KS*(u-EPI_US)))/2.) - s)*ts;
	       //Update gates

           v += dv*dt;
           w += dw*dt;
           s += ds*dt;

	       //Compute currents
           jfi = (u  > EPI_THV && v > 0.0)  ? -v  * (u - EPI_THV) * (EPI_UU - u) * ONE_O_EPI_TFI : 0.0;
           jso = (u  > EPI_THSO) ? tso : u * to;
           jsi = (u  > EPI_THSI) ? -w * s * ONE_O_EPI_TSI : 0.0;

           if (i >= 5){
    	      stim = 0.0;
           }
           cable[i].u = u - (jfi+jso+jsi-stim)*dt + lap[i];
           cable[i].v = v;
           cable[i].w = w;
           cable[i].s = s;
	  }
}


__global__ void calc_lap_cable (MMCell *cable, Real *lap, Real ddt_o_dx2, int nx){
	int i = blockIdx.x * blockDim.x + threadIdx.x;
	
	if (i == 0) {
	    lap[i] = ddt_o_dx2 * (-2.*cable[0].u + 2.*cable[1].u);
	}else if (i == nx-1){
	    lap[i] = ddt_o_dx2 * (-2.*cable[nx-1].u +2.*cable[nx-2].u);
	}else{
	    lap[i] = ddt_o_dx2 * (cable[i+1].u + cable[i-1].u - 2.0 * cable[i].u);
	}
	
}






void writeBinOutput2 (char * filename, MMCell * cable_cells, int nx){
	 FILE * output1;

	 output1 = fopen(filename, "ab");

	 if (output1 == NULL){
	          printf("Invalid file name\n");
	          exit(1);
	 }else{
	          fwrite (cable_cells, sizeof(MMCell), nx, output1);
	          fclose(output1);
     }
}


void initOutput2 (char * filename, MMCell * cable_cells, int nx){
	 FILE * output1;

	 output1 = fopen(filename, "w");

	 if (output1 == NULL){
         printf("Invalid file name\n");
         exit(1);
     }else{
         fwrite (cable_cells, sizeof(MMCell), nx, output1);
         fclose(output1);
	 }
}
void init_cable (MMCell *cable, Real ui, Real vi, Real wi, Real si, int nx){
	 int i;
	 for (i=0; i < nx; i++){
		  cable[i].u  = ui;
		  cable[i].v  = vi;
		  cable[i].w  = wi;
		  cable[i].s  = si;
	 }
}

//This code simulate a single cell.
int main(int argc, char ** argv) {
	 Real *lap_gpu;
	 MMCell *cable_cpu;
	 
	 clock_t c1, c2;
	 MMCell *cable_gpu;
     
     int nx        = 1024;   //Number of cells in the cable (the first and the last are artifacts)
                             //In CUDA this should be multiple of 16
     Real dt       = 0.02;   //Integration step in ms
     Real diff     = 0.001;  //Diffusion constant
     Real dx       = 0.025;  //Distance between two cells
     Real out_time = 1.0;    //Output time in ms

     Real istimmag = 0.66;   //Stimulus intensity 
     Real stimdur  = 1.0;    //Stimulus duration

     Real ddt_o_dx2 = dt*diff/(dx*dx); //Calc the diffusion term once for all
     Real cycle     = 200;

     char * file_name = (char *) calloc (50,      sizeof(char));
     Real sim_time = 10000.0; //in ms
     int imod    = (int)floor(out_time/dt + 0.5); //Number of iteration before print the results again
     int icycle  = (int)floor(cycle/dt + 0.5);
     int istimdur  = (int)floor(stimdur/dt);
     int ncycles = (int)floor(sim_time/cycle);
     if (ncycles < (sim_time/cycle)) {
         ncycles=ncycles+1;
     }

     int nend=icycle*ncycles;

     

     /***************************************/
     /*           Memory allocation         */
     /***************************************/
     
     /*Memory Allocation on the CPU side*/
     
     cable_cpu   = (MMCell *) calloc (nx, sizeof(MMCell));
     HANDLE_ERROR(cudaMalloc((void**) &cable_gpu,          (nx) * sizeof(MMCell)));
     HANDLE_ERROR(cudaMalloc((void**) &lap_gpu,            (nx) * sizeof(Real)));  
     
     /***************************************/
     /*          Inizialization             */
     /***************************************/
     
     init_cable (cable_cpu, 0.0, 1.0, 1.0, 0.0, nx);
     

     HANDLE_ERROR(cudaMemcpy(cable_gpu, cable_cpu, nx * sizeof(MMCell), cudaMemcpyHostToDevice));

     /*Definition of Block Grid and Block Size*/
     dim3 dimGrid, dimBlock;
     dimGrid  = dim3(nx/256, 1, 1);
     dimBlock = dim3(   256, 1, 1);
     
     /***************************************/
     /*          Simulation                 */
     /***************************************/
    c1 = clock();
    for (int ntime=0; ntime <= nend; ntime++){
        /****************************************/
        /* Calculation of the diffusion term    */
        /* using CUDA                           */
        /****************************************/
    	calc_lap_cable<<<dimGrid, dimBlock>>>  (cable_gpu, lap_gpu, ddt_o_dx2, nx);

    	
        /*********************************************/
        /* End of calculation of the diffusion term  */
        /*********************************************/

    	/* Output data */
    	if (ntime == 0){ //I initialize the files in which I'm going to write
    		 printf("Files inizialization\n");
    		HANDLE_ERROR(cudaMemcpy(cable_cpu, cable_gpu, nx * sizeof(MMCell), cudaMemcpyDeviceToHost));
    		sprintf(file_name, "cable.dat");
    		initOutput2 (file_name, cable_cpu, nx);
    		//printf("****** writing ntime=%d %f\n", ntime, ntime*dt);

    	}else if ((ntime % imod) == 0){
    		HANDLE_ERROR(cudaMemcpy(cable_cpu, cable_gpu, nx * sizeof(MMCell), cudaMemcpyDeviceToHost));
    		sprintf(file_name, "cable.dat");
    		writeBinOutput2 (file_name, cable_cpu, nx);
    		//printf("****** writing ntime=%d %f\n", ntime, ntime*dt);
    	 }
    	if (((ntime % icycle) < istimdur)){
    	   cstep_epi_gpu_aligned <<<dimGrid, dimBlock>>>( cable_gpu, nx, dt, lap_gpu, istimmag);
    	}else{
    	   cstep_epi_gpu_aligned <<<dimGrid, dimBlock>>>( cable_gpu, nx, dt, lap_gpu, 0.0);
    	}
    	
     }
    cutilSafeCall( cudaThreadSynchronize() );
    c2 = clock();
    double time_exec = c2 - c1;
    printf("Execution time in CUDA aligned:  %.4f msec\n", time_exec/ (CLOCKS_PER_SEC));
    return 0;
}


