/***************************************************************/
/*     Cable Exercise for CMACS workshop                       */
/*                                                             */
/*                                                             */
/*     Author: E. Bartocci (eziobart@ams.sunysb.edu)           */
/*                                                             */
/*     Description:                                            */
/*                                                             */
/*       Serial Code for a cable                               */
/*                                                             */
/*     Date:  11/29/10                                         */
/*                                                             */
/*   Free distribution with authors permission                 */
/*                                                             */
/* ==   SUNY Stony Brook, Stony Brook, NY                      */
/* ==                                                          */
/* ============================================================*/


/*The following are the parameters that you can find in the paper
 * A. Bueno-Orovio, M. Cherry, and F. Fenton, Minimal model for
 * human ventricular action potentials in tissue, Journal of
 * Theoretical Biology, no. 253, pp. 544560, 2008.              */

#define EPI_TVP        1.4506
#define ONE_O_EPI_TVP  0.68936991589687

#define EPI_TV1M     60.
#define EPI_TV2M     1150.

#define ONE_O_EPI_TV1M 0.016666666666667
#define ONE_O_EPI_TV2M 0.000869565217391

#define EPI_TWP      200.0

#define ONE_O_EPI_TWP  0.005

#define EPI_TW1M     60.0
#define EPI_TW2M     15.     //The same with Flavio's paper

#define EPI_TS1      2.7342  //The same with Flavio's paper
#define EPI_TS2      16.     //The same with Flavio's paper

#define ONE_O_EPI_TS1 0.365737692926633
#define ONE_O_EPI_TS2 0.0625

#define EPI_TFI      0.11    //The same with Flavio's paper
#define ONE_O_EPI_TFI 9.090909090909091

#define EPI_TO1      400.    //The same with Flavio's paper
#define EPI_TO2      6.      //The same with Flavio's paper

#define ONE_O_EPI_TO1 0.0025
#define ONE_O_EPI_TO2 0.166666666666667

#define ONE_O_EPI_TO1 0.0025
#define ONE_O_EPI_TO2 0.166666666666667

#define EPI_TSO1     30.0181 //The same with Flavio's paper
#define EPI_TSO2     0.9957  //The same with Flavio's paper

#define EPI_TSI      1.8875
#define ONE_O_EPI_TSI 0.529801324503311

#define EPI_TWINF    0.07
#define ONE_O_EPI_TWINF    14.285714285714286

#define EPI_THV      0.3
#define EPI_THVM     0.006
#define EPI_THVINF   0.006
#define EPI_THW      0.13
#define EPI_THWINF   0.006
#define EPI_THSO     0.13
#define EPI_THSI     0.13
#define EPI_THO      0.006
#define EPI_KWM      65.     //The same of Flavio's paper
#define EPI_KS       2.0994  //The same of Flavio's paper
#define EPI_KSO      2.0458  //The same of Flavio's paper
#define EPI_UWM      0.03    //The same of Flavio's paper
#define EPI_US       0.9087  //The same of Flavio's paper
#define EPI_UO       0.     // The same of Flavio's paper
#define EPI_UU       1.55   // The same of Flavio's paper
#define EPI_USO      0.65   // The same of Flavio's paper
#define EPI_SC       0.007

#define EPI_WINFSTAR 0.94   // The same of Flavio's paper

#define EPI_TW2M_TW1M -45.0
#define EPI_TSO2_TSO1 -29.0224

#define EPI_TW2M_TW1M_DIVBY2 -22.5
#define EPI_TSO2_TSO1_DIVBY2 -14.5112


typedef float Real; /*To use double we need a NVIDIA card
                      with CUDA compute capability 1.3*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>


void writeBinOutput (char * filename, Real * state_vars, int nx){
	 FILE * output1;

	 output1 = fopen(filename, "ab");

	 if (output1 == NULL){
         printf("Invalid file name\n");
         exit(1);
     }else{
         fwrite (state_vars, sizeof(Real), nx, output1);
         fclose(output1);
	 }
}


void initOutput (char * filename, Real * state_vars, int nx){
	 FILE * output1;

	 output1 = fopen(filename, "w");

	 if (output1 == NULL){
         printf("Invalid file name\n");
         exit(1);
     }else{
         fwrite (state_vars, sizeof(Real), nx, output1);
         fclose(output1);
	 }
}


void initialize (Real * u, Real * v, Real * w, Real *s, Real ui, Real vi, Real wi, Real si, int nx){
	 int i;
	 for (i=0; i < nx; i++){
		  u[i]  = ui;
		  v[i]  = vi;
		  w[i]  = wi;
		  s[i]  = si;
	 }
}

void calc_lap_cpu (Real *u, Real *lap, Real ddt_o_dx2, int nx){
	int i;
	for (i=0; i < nx; i++){
	     if (i == 0) {
	         lap[i] = ddt_o_dx2 * (-2.*u[0] + 2.*u[1]);
	     }else if (i == nx-1){
	         lap[i] = ddt_o_dx2 * (-2.*u[nx-1] +2.*u[nx-2]);
	     }else{
	         lap[i] = ddt_o_dx2 * (u[i+1] + u[i-1] - 2.0 * u[i]);
	     }
	}
}

void  cstep_epi ( Real *ui, Real *vi, Real *wi, Real *si,  Real dt, Real lap, Real stim){

	  Real tvm, vinf, winf;
	  Real jfi, jso, jsi;
	  Real twm, tso, ts, to, ds, dw, dv;

	  tvm =  (*ui > EPI_THVM) ? EPI_TV2M : EPI_TV1M;
      twm =  EPI_TW1M + (EPI_TW2M_TW1M_DIVBY2)*(1.+tanh(EPI_KWM*(*ui-EPI_UWM)));
      tso =  EPI_TSO1 + (EPI_TSO2_TSO1_DIVBY2)*(1.+tanh(EPI_KSO*(*ui-EPI_USO)));
      ts  = (*ui > EPI_THW)  ? EPI_TS2  : EPI_TS1;
      to  = (*ui > EPI_THO)  ? EPI_TO2  : EPI_TO1;

      vinf = (*ui >  EPI_THVINF) ? 0.0: 1.0;
	  winf = (*ui >  EPI_THWINF) ? EPI_WINFSTAR: 1.0-*ui/EPI_TWINF;
      if (winf > 1) winf = 1.0;

      dv = (*ui > EPI_THV) ? -*vi/EPI_TVP : (vinf-*vi)/tvm;
      dw = (*ui > EPI_THW) ? -*wi/EPI_TWP : (winf-*wi)/twm;
      ds = (((1.+tanh(EPI_KS*(*ui-EPI_US)))/2.) - *si)/ts;

	  //Update gates

       *vi += dv*dt;
       *wi += dw*dt;
       *si += ds*dt;

	  //Compute currents
      jfi = (*ui > EPI_THV)  ? -*vi * (*ui - EPI_THV) * (EPI_UU - *ui)/EPI_TFI : 0.0;
      jso = (*ui > EPI_THSO) ? 1./tso : (*ui-EPI_UO)/to;
      jsi = (*ui > EPI_THSI) ? -*wi * *si/EPI_TSI : 0.0;

      *ui = *ui  - (jfi+jso+jsi-stim)*dt + lap;
}



//This code simulate a single cell.
int main(int argc, char ** argv) {
	 Real *state_vars, *u, *v, *w, *s, *lap_cpu;

	 clock_t c1, c2;
     int ntime, i;
     int nx        = 1024;   //Number of cells in the cable (the first and the last are artifacts)
                             //In CUDA this should be multiple of 16
     Real dt       = 0.02;   //Integration step in ms
     Real diff     = 0.001;  //Diffusion constant
     Real dx       = 0.025;  //Distance between two cells
     Real out_time = 1.0;    //Output time in ms

     Real istimmag = 0.66;   //Stimulus intensity
     Real stimdur  = 1.0;    //Stimulus duration

     Real ddt_o_dx2 = dt*diff/(dx*dx); //Calc the diffusion term once for all
     Real cycle     = 200;

     char * file_name = (char *) calloc (50,      sizeof(char));
     Real sim_time = 10000.0; //in ms
     int imod    = (int)floor(out_time/dt + 0.5); //Number of iteration before print the results again
     int icycle  = (int)floor(cycle/dt + 0.5);
     int istimdur  = (int)floor(stimdur/dt);
     int ncycles = (int)floor(sim_time/cycle);
     if (ncycles < (sim_time/cycle)) {
         ncycles=ncycles+1;
     }

     int nend=icycle*ncycles;



     /***************************************/
     /*           Memory allocation         */
     /***************************************/

     /*Memory Allocation on the CPU side*/
     state_vars  = (Real *) calloc (4 * nx, sizeof (Real));
     lap_cpu     = (Real *) calloc   (    nx, sizeof (Real));

     /*Assignment of the pointers*/
     u       = state_vars;
     v       = state_vars +      nx;
     w       = state_vars + 2 *  nx;
     s       = state_vars + 3 *  nx;
     /***************************************/
     /*          Inizialization             */
     /***************************************/
     initialize(u, v, w, s, 0.0, 1.0, 1.0, 0.0, nx);

     /***************************************/
     /*          Simulation                 */
     /***************************************/
    c1 = clock();
    for (ntime=0; ntime <= nend; ntime++){
        /****************************************/
        /* Calculation of the diffusion term    */
        /* using CUDA                           */
        /****************************************/

    	calc_lap_cpu(state_vars, lap_cpu, ddt_o_dx2, nx);

        /*********************************************/
        /* End of calculation of the diffusion term  */
        /*********************************************/

    	/* Output data */
    	if (ntime == 0){ //I initialize the files in which I'm going to write
    		 printf("Files inizialization\n");


    	}else if ((ntime % imod) == 0){
    	     //printf("****** writing ntime=%d %f\n", ntime, ntime*dt);
    	     sprintf(file_name, "result_mfa_u_cycle_%d_nx_%d.dat", (int)cycle, nx);
    	     writeBinOutput (file_name, u, nx);
    	     sprintf(file_name, "result_mfa_v_cycle_%d_nx_%d.dat", (int)cycle, nx);
    	     writeBinOutput (file_name, v, nx);
    	     sprintf(file_name, "result_mfa_w_cycle_%d_nx_%d.dat", (int)cycle, nx);
    	     writeBinOutput (file_name, w, nx);
    	     sprintf(file_name, "result_mfa_s_cycle_%d_nx_%d.dat", (int)cycle, nx);
    	     writeBinOutput (file_name, s, nx);
    	 }

    	for (i=0; i < nx; i++){
    	    if (((ntime % icycle) < istimdur) && (i< 5)){
    	    	cstep_epi (&u[i], &v[i], &w[i], &s[i],  dt, lap_cpu[i], istimmag);

    	    }else{
    	    	cstep_epi (&u[i], &v[i], &w[i], &s[i],  dt, lap_cpu[i], 0.0);
    	    }

    	}

     }

    c2 = clock();
    double time_exec = c2 - c1;

    printf("Execution time in the serial: %.4f msec\n", time_exec/ (CLOCKS_PER_SEC));

    return 0;
}
